/*
 *			P I N G . C
 *
 * Using the Internet Control Message Protocol (ICMP) "ECHO" facility,
 * measure round-trip-delays and packet loss across network paths.
 *
 * Author -
 *	Mike Muuss
 *	U. S. Army Ballistic Research Laboratory
 *	December, 1983
 * Modified at Uc Berkeley
 * Modified at Rutgers: Ron Natalie, David Paul Zimmerman
 * Modified at DECWRL: Jeffrey Mogul
 * Record Route and verbose headers - Phil Dykstra, BRL, March 1988.
 * Modified at Cornell: Jeffrey C Honig, April 1989
 *
 * Status -
 *	Public Domain.  Distribution Unlimited.
 *
 * Bugs -
 *	More statistics could always be gathered.
 *	This program has to run SUID to ROOT to access the ICMP socket.
 */

/*
 * Extensively modified by Eric Wassenaar, Nikhef-H, <e07@nikhef.nl>
 *
 * The source of this particular version of the program is available
 * via anonymous ftp from machine 'ftp.nikhef.nl' [192.16.199.1]
 * in the directory '/pub/network' as 'ping.tar.Z'
 *
 * Bugs fixed
 *
 *	Checksum algorithm.
 *		The ICMP packet checksum was computed incorrectly
 *		in case the packet contained an odd number of bytes.
 *	Hostname storage.
 *		The hostname as returned via gethostby{name,addr} was
 *		pointing to internal storage which would be clobbered
 *		by subsequent calls. Furthermore, MAXHOSTNAMELEN is
 *		not sufficient to store fully qualified domain names.
 *	Response printout.
 *		The printing of the response messages was not always
 *		according to the specified options.
 *	Event scheduling.
 *		The scheduling of new events and termination was not
 *		always properly corresponding to the specified options.
 *	Packet loss report.
 *		Packet loss is now properly reported. Especially when
 *		running in flood mode the figures were bogus.
 *
 * New features
 *
 *	Internal redesign.
 *		Necessary to properly handle flood mode and broadcast.
 *	Portability hooks.
 *		New Makefile and header files to ease installation on
 *		various platforms.
 *	Loose source route.
 *		Can be specified instead of, or in combination with,
 *		record route. Handling of IP options is generalized.
 *		See conf.h for potential conflicts with record route.
 *	Broadcast support.
 *		Added hooks to handle broadcast address in all modes.
 *		Restrict the use of flood ping to the superuser when
 *		a broadcast address is specified.
 *	Quick ping.
 *		Suppresses all output and quits when target is alive.
 *
 * Released versions
 *
 *	910926
 *		getopt() returns int instead of char.
 *		Put omninet (NC400) controller in promiscuous mode.
 *		This is necessary for record route mode to work.
 *	940212
 *		Use setvbuf() if setlinebuf() is unavailable.
 *		Use float format to print percentage packet loss.
 *		Miscellaneous lint cleanup.
 *		Save hostname properly in sufficient static storage.
 *		Correct checksum algorithm for odd bytecount.
 *		Suppress output in quiet flood mode.
 *	940217
 *		Complete overhaul of entire code. Various bugs fixed.
 *		General portability hooks for various platforms.
 *		Improved error reporting and timeout handling.
 *		Corrected output print format depending on options.
 *		Properly schedule next event in various modes.
 *		More robust code in several places.
 *		Undocumented option -T to set timeout in flood mode.
 *		Redesign flow control to properly handle flood mode.
 *		This ensures that packet loss figures are now reliable.
 *	940226
 *		Generalize ip options processing.
 *		New option -L to enable loose source route mode.
 *		Recognize option -l for compatibility with SUN ping.
 *		Restrict simultaneous use of more than one IP routing
 *		option, unless allowed via an explicit compile switch.
 *		Make messages during verbose packet reporting uniform.
 *		Implement special processing for broadcast pings.
 *		Restrict options if a broadcast address is specified.
 *		Properly pluralize words in output messages.
 *		New option -Q for quick ping without statistics.
 *		Manual page revised and adapted to new features.
 *	940306
 *		Word-align all 32-bit addresses in IP options buffer.
 *		Check class D (multicast) and class E (reserved) address.
 *		Keep track of different host responses in broadcast mode.
 *
 *	940704
 *		General portability fixes.
 *		Use predefined constants for fixed size entities.
 *		Cannot rely on sizeof() for several structs on those
 *		platforms which don't have 16 or 32 bit integers.
 *		Adapt checksum algorithm to be more portable.
 */

/*
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char Version[] = "@(#)ping.c	e07@nikhef.nl (Eric Wassenaar) 940704";
#endif

#if defined(apollo) && defined(lint)
#define __attribute(x)
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#undef NOERROR			/* in <sys/streams.h> on solaris 2.x */
#include <arpa/nameser.h>
#include <resolv.h>

#include "conf.h"		/* various configuration definitions */
#include "exit.h"		/* exit codes come from <sysexits.h> */
#include "port.h"		/* various portability definitions */

typedef int	bool;		/* boolean type */
#define TRUE	1
#define FALSE	0

#define STDIN	0
#define STDOUT	1
#define STDERR	2

#ifdef lint
#define EXTERN
#else
#define EXTERN extern
#endif

EXTERN int errno;

extern char *version;		/* program version number */

/*
 * Various buffer sizes.
 */

#ifndef MAXPKT
#define	MAXPKT	4096			/* max output packet total size */
#endif
#define PKTSIZE	64			/* default output packet size */
#define HDRLEN	ICMP_MINLEN		/* icmp header length (8) */
#define	TIMLEN	sizeof(struct timeval)	/* size of timer data (8) */
#define DATALEN	(PKTSIZE-HDRLEN)	/* default packet data length (56) */
#define	MAXDATA	(MAXPKT-HDRLEN-TIMLEN)	/* max available for fill data */

#ifdef IP_MAXPACKET
#define MAXPACKET IP_MAXPACKET	/* max ip packet size */
#else
#define MAXPACKET 65535
#endif

#define IPHDRSZ		20	/* actually sizeof(struct ip) */

#ifdef IP_OPTIONS
#ifndef MAX_IPOPTLEN
#define MAX_IPOPTLEN	40	/* max ip options buffer size */
#endif
#define IPOPT_HDRLEN	3	/* actually IPOPT_MINOFF - 1 */
#endif /*IP_OPTIONS*/

#define	RCVBUF	48*1024		/* size of receive buffer to specify */

#define	MAXWAIT	5		/* max secs to wait for response */

/*
 * Modes of operation.
 */

#define PING_NORMAL	0	/* ping at regular intervals */
#define PING_FLOOD	1	/* ping as fast as possible */
#define PING_CISCO	2	/* ping Cisco-style */

int pingmode = PING_NORMAL;	/* how to ping */

/*
 * Command line flags.
 */

int verbose = 0;		/* -v  print additional information */
bool quiet = FALSE;		/* -q  don't print markers, only summary */
bool quick = FALSE;		/* -Q  don't print summary, quit if alive */
bool numeric = FALSE;		/* -n  print IP address as dotted quad */
bool fastping = FALSE;		/* -F  next ping immediately upon response */
bool printmiss = FALSE;		/* -m  print missed responses */
bool traceroute = FALSE;	/* -R  enable packet route recording */
bool looseroute = FALSE;	/* -L  enable loose source routing */

/*
 * Command line options.
 */

int preload = 0;		/* -p  number of packets to "preload" */
int timeout = 1;		/* -t  timeout between packets (secs) */
int floodtime = 10;		/* -T  timeout in flood mode (millisecs) */
int filldata = 0;		/* -D  pattern for data packet specified */

int datalen = DATALEN;		/* size of probe packet */
int npackets = 0;		/* maximum number of packets to send */

char hostnamebuf[MAXDNAME+1];
char *hostname;			/* target host name */

/*
 * Internal control variables.
 */

bool timing = TRUE;		/* set if packet contains timing info */
bool gotone = FALSE;		/* set if we got a valid reponse */
bool flushing = FALSE;		/* set if shutdown initiated */
bool broadcast = FALSE;		/* set if broadcast address specified */
int column = 0;			/* column number for cisco pings */
int ident = 0;			/* packet identifier */
int ntransmitted = 0;		/* seq # for outbound packets */

int request = ICMP_ECHO;	/* outbound icmp request type */
int reply = ICMP_ECHOREPLY;	/* expected inbound reply type */

int sock;			/* socket file descriptor */
int sockopts = 0;		/* socket options */

struct sockaddr myaddr;		/* address of ourselves */
struct sockaddr toaddr;		/* address to send to */
struct sockaddr fromaddr;	/* address to recv from */

struct sockaddr_in *me   = (struct sockaddr_in *)&myaddr;
struct sockaddr_in *to   = (struct sockaddr_in *)&toaddr;
struct sockaddr_in *from = (struct sockaddr_in *)&fromaddr;

u_char opacket[MAXPACKET];	/* outgoing packet */
u_char ipacket[MAXPACKET];	/* incoming packet */

/*
 * BITMAPSIZE is the number of bits in received bitmap, i.e. the
 * maximum number of received sequence numbers we can keep track of.
 * Use 2048 for complete accuracy -- sequence numbers are 16 bits.
 */

#define WORDSIZE	(8 * sizeof(u_int))
#define BITMAPSIZE	(WORDSIZE * 2048)

u_int rcvd_bitmap[BITMAPSIZE / WORDSIZE];

#define MAPBIT(bit)	((bit) % BITMAPSIZE)		/* bit in bitmap */
#define MAPWORD(bit)	(MAPBIT(bit) / WORDSIZE)	/* word in bitmap */
#define WORDBIT(bit)	(1 << (MAPBIT(bit) % WORDSIZE))	/* bit in word */

#define SET(bit, map)	(map)[MAPWORD(bit)] |= WORDBIT(bit)
#define CLR(bit, map)	(map)[MAPWORD(bit)] &= ~WORDBIT(bit)
#define TST(bit, map)	(((map)[MAPWORD(bit)] & WORDBIT(bit)) != 0)

#define SETBIT(bit)	SET(bit, rcvd_bitmap)
#define CLRBIT(bit)	CLR(bit, rcvd_bitmap)
#define TSTBIT(bit)	TST(bit, rcvd_bitmap)

/*
 * Structure for statistics on round-trip times.
 */

typedef struct _stats
{
	int rcvd;			/* number of responses */
	int dupl;			/* number of duplicates */
	time_t rttmin;			/* minimum round-trip time */
	time_t rttmax;			/* maximum round-trip time */
	double rttsum;			/* sum of recorded rtt values */
} stats;

#define VERY_LONG ((time_t)999999999)

stats pingstats = { 0, 0, VERY_LONG, 0, 0.0 };

/*
 * Structure for host info in broadcast mode.
 */

typedef struct _hostinfo
{
	struct _hostinfo *next;		/* next in chain */
	struct in_addr inaddr;		/* IP address */
	int rcvd;			/* number of responses */
} hostinfo;

hostinfo *hostchain = NULL;	/* chain of recorded hosts */

/*
 * Structure for storing responses for a specific route.
 */

#ifdef IP_OPTIONS

typedef struct _optstr
{
	struct _optstr *next;		/* next in chain */
	struct _stats stats;		/* routing statistics */
	u_char ipopt[MAX_IPOPTLEN];	/* routing options */
	int ipoptlen;			/* actual size of options */
} optstr;

optstr *optchain = NULL;	/* chain of recorded routes */

#endif /*IP_OPTIONS*/

/*
 * Useful inline functions.
 */

#define superuser()	(getuid() == 0)

#define bitset(bit, w)	(((w) & (bit)) != 0)

#define	atox(c)		(isdigit(c) ? (c - '0')      : \
			(isupper(c) ? (c - 'A' + 10) : \
			(islower(c) ? (c - 'a' + 10) : 0)))

#define plural(n)	((n) == 1 ? "" : "s")

#define sameaddr(a,b)	((a)->sin_addr.s_addr == (b)->sin_addr.s_addr)

#define NOT_DOTTED_QUAD	((ipaddr_t)-1)

#define ANY_ADDR	((ipaddr_t)0x00000000)
#define BROADCAST_ADDR	((ipaddr_t)0xFFFFFFFF)

#include "defs.h"	/* declaration of functions */

#ifdef DEBUG
#define assert(condition)\
{\
	if (!(condition))\
	{\
		(void) fprintf(stderr, "assertion botch: ");\
		(void) fprintf(stderr, "%s(%d): ", __FILE__, __LINE__);\
		(void) fprintf(stderr, "%s\n", "condition");\
		exit(EX_SOFTWARE);\
	}\
}
#else
#define assert(condition)
#endif

static char Usage[] =
"\
Usage:   %s [options] host [length [npackets]]\n\
Flags:   [-c|-f] [-F] [-Q] [-LR] [-mnqv] [-dr]\n\
Options: [-p preload] [-t timeout] [-D pattern]\
";

/*
** MAIN -- Start of program ping
** -----------------------------
**
**	Normal exits are via finish(), when the maximum packet count
**	is exhausted or when terminated with an interrupt.
**	Premature exits occur when abnormal conditions are detected.
**
**	Exit status:
**		Various possibilities from <sysexits.h>
**		EX_OK if at least one valid response was received.
*/

int
main(argc, argv)
int argc;
char *argv[];
{
	register char *option;
	register char *cp;
	register int i;
	char *program;			/* name that ping was called with */
	ipaddr_t addr;
	struct in_addr inaddr;
	struct hostent *hp;
	u_char *pattern = &opacket[HDRLEN+TIMLEN];

#ifdef obsolete
	assert(sizeof(u_int) == 4);	/* probably paranoid */
	assert(sizeof(u_short) == 2);	/* perhaps less paranoid */
	assert(sizeof(ipaddr_t) == 4);	/* but this is critical */
#endif /*obsolete*/

/*
 * Synchronize stdout and stderr in case output is redirected.
 */
	linebufmode(stdout);

/*
 * Pickup our own address.
 */
	if (gethostname(hostnamebuf, MAXDNAME) < 0)
	{
		perror("gethostname");
		exit(EX_OSERR);
	}
	hostnamebuf[MAXDNAME] = '\0';

	hp = gethostbyname(hostnamebuf);
	if (hp == NULL)
	{
		error("Cannot find own name %s", hostnamebuf);
		exit(EX_NOHOST);
	}
	bcopy(hp->h_addr, (char *)&inaddr, INADDRSZ);

	/* setup ip address */
	bzero((char *)&myaddr, sizeof(myaddr));
	me->sin_family = AF_INET;
	me->sin_addr = inaddr;
	me->sin_port = 0;

/*
 * Scan command line options and flags.
 */
	program = rindex(argv[0], '/');
	if (program++ == NULL)
		program = argv[0];

	while (argc > 1 && argv[1][0] == '-')
	{
	    for (option = &argv[1][1]; *option != '\0'; option++)
	    {
		switch (*option)
		{
		    case 'd':		/* socket debugging */
			sockopts |= SO_DEBUG;
			break;

		    case 'r':		/* don't use routing table */
			sockopts |= SO_DONTROUTE;
			break;

		    case 'm':		/* show missing response stats */
			printmiss = TRUE;
			break;

		    case 'n':		/* numeric IP addresses only */
			numeric = TRUE;
			break;

		    case 'Q':		/* don't print summary */
			quick = TRUE;
			/* fall through */

		    case 'q':		/* don't print markers */
			quiet = TRUE;
			break;

		    case 'v':		/* increment verbosity level */
			verbose++;
			break;

		    case 'l':		/* compat with older versions */
			/* fall through */

		    case 'L':		/* loose source route */
			looseroute = TRUE;
			break;

		    case 'R':		/* record route */
			traceroute = TRUE;
			break;

		    case 'F':		/* no delay between pings */
			fastping = TRUE;
			break;

		    case 'f':		/* flood pings */
			if (pingmode == PING_CISCO)
				fatal("Conflicting options -f and -c");
			pingmode = PING_FLOOD;
			break;

		    case 'c':		/* cisco style pings */
			if (pingmode == PING_FLOOD)
				fatal("Conflicting options -c and -f");
			pingmode = PING_CISCO;
			break;

		    case 'p':		/* # of packets to preload */
			if (argv[2] == NULL || argv[2][0] == '-')
				fatal("Missing preload value");
			preload = atoi(argv[2]);
			if (preload < 1)
				fatal("Invalid preload value %s", argv[2]);
			argv++; argc--;
			break;

		    case 't':		/* timeout between pings */
			if (argv[2] == NULL || argv[2][0] == '-')
				fatal("Missing timeout value");
			timeout = atoi(argv[2]);
			if (timeout < 1)
				fatal("Invalid timeout value %s", argv[2]);
			argv++; argc--;
			break;

		    case 'T':		/* timeout for flood mode */
			if (argv[2] == NULL || argv[2][0] == '-')
				fatal("Missing timeout value");
			floodtime = atoi(argv[2]);
			if (floodtime < 1)
				fatal("Invalid timeout value %s", argv[2]);
			if (floodtime > 1000)
				fatal("Maximum timeout value %s", itoa(1000));
			argv++; argc--;
			break;

		    case 'D':		/* specify fill data */
			if (argv[2] == NULL || argv[2][0] == '-')
				fatal("Missing pattern data");
			for (cp = argv[2]; *cp != '\0'; cp += 2)
			{
				u_char pat;

				if (!isxdigit(cp[0]) || !isxdigit(cp[1]))
					fatal("Invalid hex data %s", argv[2]);
				if (filldata >= MAXDATA)
					fatal("Too much fill data specified");
				pat = (atox(cp[0]) << 4) | atox(cp[1]);
				pattern[filldata++] = pat;
			}
			argv++; argc--;
			break;

		    case 's':		/* compat with older versions */
			break;

		    case 'V' :
			printf("Version %s\n", version);
			exit(EX_OK);

		    default:
			fatal(Usage, program);
			/*NOTREACHED*/
		}
	    }

	    argv++; argc--;
	}

/*
 * Check availability of special routing options.
 */
#ifdef IP_OPTIONS

#ifndef MULTIPLE_IP_OPTIONS
	if ((traceroute && looseroute) && !superuser())
		fatal("Conflicting options -R and -L");
#endif

#else /*IP_OPTIONS*/

	if (traceroute)
		error("record route not supported");
	if (looseroute)
		error("loose source route not supported");

#endif /*IP_OPTIONS*/

/*
 * First argument is mandatory.
 */
	if (argc < 2)
		fatal(Usage, program);

	/* hostname or dotted quad to ping */
	hostname = argv[1];
	addr = inet_addr(hostname);
	inaddr.s_addr = addr;

	if (addr != NOT_DOTTED_QUAD)
	{
		hp = gethostbyaddr((char *)&inaddr, INADDRSZ, AF_INET);
		if (hp == NULL)
			hostname = strcpy(hostnamebuf, hostname);
		else
			hostname = strcpy(hostnamebuf, hp->h_name);
	}
	else
	{
		hp = gethostbyname(hostname);
		if (hp == NULL)
		{
			error("Unknown host %s", hostname);
			exit(EX_NOHOST);
		}

		hostname = strcpy(hostnamebuf, hp->h_name);
		bcopy(hp->h_addr, (char *)&inaddr, INADDRSZ);
	}

	/* setup ip address */
	bzero((char *)&toaddr, sizeof(toaddr));
	to->sin_family = AF_INET;
	to->sin_addr = inaddr;
	to->sin_port = 0;

	/* special handling for broadcast address */
	broadcast = bcast_addr(to->sin_addr);

	/* this is one of the worst network killers ... */
	if ((broadcast && pingmode == PING_FLOOD) && !superuser())
		fatal("No flood ping to broadcast address");

/*
 * Scan remaining optional command line arguments.
 */
	/* data length is packet size minus header length */
	if (argc > 2)
	{
		datalen = atoi(argv[2]);
		if (datalen < 1)
			fatal("Invalid packet length %s", argv[2]);
		if (datalen > MAXPKT - HDRLEN)
			fatal("Maximum packet length %s", itoa(MAXPKT-HDRLEN));
	}

	/* maximum number of packets */
	if (argc > 3)
	{
		npackets = atoi(argv[3]);
		if (npackets < 1)
			fatal("Invalid packet count %s", argv[3]);
	}

	/* rest is undefined */
	if (argc > 4)
		fatal(Usage, program);

/*
 * Miscellaneous initialization.
 */
	/* our packet identifier */
	ident = getpid() & 0xFFFF;

#ifdef IP_OPTIONS
	/* here we route our own reply packet */
	if (looseroute)
		request = ICMP_ECHOREPLY;
#endif /*IP_OPTIONS*/

	/* can do timing only if packet big enough */
	timing = (datalen >= TIMLEN) ? TRUE : FALSE;

	/* stuff data portion of output packet */
	if (filldata)
	{
		/* pattern specified -- replicate to fill buffer */
		for (i = 0; i < MAXDATA - filldata; i++)
			pattern[filldata+i] = pattern[i];
	}
	else
	{
		/* no pattern specified -- fill with position info */
		for (i = 0; i < MAXDATA; i++)
			pattern[i] = TIMLEN + i;
	}

#ifdef OMNINET
	/* initialize network device for route recording to work */
	if (traceroute)
		(void) initdevice(OMNINET);
#endif /*OMNINET*/

/*
 * Allocate and configure raw icmp socket.
 */
	check_proto();

	get_socket();

#ifdef IP_OPTIONS
 	if (traceroute || looseroute)
		set_options();
#endif /*IP_OPTIONS*/

/*
 * All set. Start off.
 */
	/* don't need special privileges any more */
	(void) setuid(getuid());

	/* show data portion of output packet */
	if (filldata && (datalen > TIMLEN))
	{
		if (!quiet)
		{
			printf("PATTERN: ");
			for (i = 0; i < filldata; i++)
				printf("%02x", pattern[i] & 0xFF);
			printf("\n");
		}
	}

	ping();
	/*NOTREACHED*/
}

/*
** FATAL -- Abort program when illegal option encountered
** ------------------------------------------------------
*/

void /*VARARGS1*/
fatal(fmt, a, b, c, d)
char *fmt;				/* format of message */
char *a, *b, *c, *d;			/* optional arguments */
{
	(void) fprintf(stderr, fmt, a, b, c, d);
	(void) fprintf(stderr, "\n");
	exit(EX_USAGE);
}


void /*VARARGS1*/
error(fmt, a, b, c, d)
char *fmt;				/* format of message */
char *a, *b, *c, *d;			/* optional arguments */
{
	(void) fprintf(stderr, fmt, a, b, c, d);
	(void) fprintf(stderr, "\n");
}

/*
** CHECK_PROTO -- Check protocol numbers
** -------------------------------------
**
**	ping uses protocol numbers as defined in <netinet/in.h>.
**	Verify whether they correspond to the values in /etc/protocols.
**	This is probably rather paranoid.
*/

void
check_proto()
{
	struct protoent *proto;

	proto = getprotobyname("ip");
	if (proto == NULL)
	{
		(void) fprintf(stderr, "ip: unknown protocol\n");
		exit(EX_OSFILE);
	}
	if (proto->p_proto != IPPROTO_IP)
	{
		(void) fprintf(stderr, "ip protocol %d should be %d\n",
			proto->p_proto, IPPROTO_IP);
		exit(EX_CONFIG);
	}

	proto = getprotobyname("icmp");
	if (proto == NULL)
	{
		(void) fprintf(stderr, "icmp: unknown protocol\n");
		exit(EX_OSFILE);
	}
	if (proto->p_proto != IPPROTO_ICMP)
	{
		(void) fprintf(stderr, "icmp protocol %d should be %d\n",
			proto->p_proto, IPPROTO_ICMP);
		exit(EX_CONFIG);
	}

	proto = getprotobyname("tcp");
	if (proto == NULL)
	{
		(void) fprintf(stderr, "tcp: unknown protocol\n");
		exit(EX_OSFILE);
	}
	if (proto->p_proto != IPPROTO_TCP)
	{
		(void) fprintf(stderr, "tcp protocol %d should be %d\n",
			proto->p_proto, IPPROTO_TCP);
		exit(EX_CONFIG);
	}

	proto = getprotobyname("udp");
	if (proto == NULL)
	{
		(void) fprintf(stderr, "udp: unknown protocol\n");
		exit(EX_OSFILE);
	}
	if (proto->p_proto != IPPROTO_UDP)
	{
		(void) fprintf(stderr, "udp protocol %d should be %d\n",
			proto->p_proto, IPPROTO_UDP);
		exit(EX_CONFIG);
	}
}

/*
** GET_SOCKET -- Allocate and configure socket
** -------------------------------------------
**
**	A raw icmp socket is allocated. This can be done only by root.
**	Extra socket options are set as requested on the command line.
*/

void
get_socket()
{
	int rcvbuf = RCVBUF;
	int on = 1;

	sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (sock < 0)
	{
		perror("socket");
		exit(EX_OSERR);
	}

	if (bitset(SO_DEBUG, sockopts))
	{
		if (setsockopt(sock, SOL_SOCKET, SO_DEBUG,
			(char *)&on, sizeof(on)) < 0)
		{
			perror("setsockopt: SO_DEBUG");
			exit(EX_OSERR);
		}
	}

	if (bitset(SO_DONTROUTE, sockopts))
	{
		if (setsockopt(sock, SOL_SOCKET, SO_DONTROUTE,
			(char *)&on, sizeof(on)) < 0)
		{
			perror("setsockopt: SO_DONTROUTE");
			exit(EX_OSERR);
		}
	}

#ifdef SO_RCVBUF
	{
		if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
			(char *)&rcvbuf, sizeof(rcvbuf)) < 0)
		{
			perror("setsockopt: SO_RCVBUF");
			exit(EX_OSERR);
		}
	}
#endif /*SO_RCVBUF*/
}

/*
** SET_OPTIONS -- Initialize IP options
** ------------------------------------
**
**	Special IP options are setup as requested on the command line.
**	This will extend the IP header with the IP options buffer.
**	Not all platforms may support this, even if IP_OPTIONS exists.
*/

#ifdef IP_OPTIONS

void
set_options()
{
	u_char ipoptbuf[MAX_IPOPTLEN];	/* ip options buffer */
	u_char *ipopt = ipoptbuf;	/* current index */

	bzero((char *)ipoptbuf, sizeof(ipoptbuf));

/*
 * Start with loose source route.
 */
	if (looseroute)
	{
		ipopt[IPOPT_OPTVAL] = IPOPT_NOP;
		ipopt++;

		ipopt[IPOPT_OPTVAL] = IPOPT_LSRR;
		ipopt[IPOPT_OLEN]   = IPOPT_HDRLEN + 2*INADDRSZ;
		ipopt[IPOPT_OFFSET] = IPOPT_MINOFF;
		ipopt += IPOPT_HDRLEN;

		bcopy((char *)&to->sin_addr, (char *)ipopt, INADDRSZ);
		ipopt += INADDRSZ;
		bcopy((char *)&me->sin_addr, (char *)ipopt, INADDRSZ);
		ipopt += INADDRSZ;
	}

/*
 * Use remainder for route recording.
 */
	if (traceroute)
	{
		ipopt[IPOPT_OPTVAL] = IPOPT_NOP;
		ipopt++;

		ipopt[IPOPT_OPTVAL] = IPOPT_RR;
		ipopt[IPOPT_OLEN]   = MAX_IPOPTLEN - (ipopt-ipoptbuf);
		ipopt[IPOPT_OFFSET] = IPOPT_MINOFF;
	}

/*
 * Request use of IP options.
 */
	if (setsockopt(sock, IPPROTO_IP, IP_OPTIONS,
		(char *)ipoptbuf, sizeof(ipoptbuf)) < 0)
	{
		perror("setsockopt: IP_OPTIONS");
		exit(EX_OSERR);
	}
}

#endif /*IP_OPTIONS*/

/*
** PING -- Main loop of ping
** -------------------------
**
**	This is the core module of the program. Initialization is done.
**	Start pinging according to the requested mode.
*/

void
ping()
{
	register int i;
	int cc;

	if (!quick)
	{
		printf("PING %s: ", pr_addr(to->sin_addr));
		printf("%d data byte%s", datalen, plural(datalen));
		if (npackets > 0)
			printf(", %d packet%s", npackets, plural(npackets));
		if (broadcast)
			printf(", broadcast");
		printf("\n");
	}

	/* setup for first interrupt */
	(void) signal(SIGINT, prefinish);

	/* force proper mode */
	if (pingmode == PING_CISCO)
		fastping = TRUE;
	if (pingmode == PING_FLOOD)
		fastping = FALSE;

	/* enforce hard limit for consistency */
	if (npackets > 0 && preload > npackets)
		preload = npackets;

	/* fire off them quickies */
	for (i = 0; i < preload; i++)
		send_ping();

	/* start things going */
	if (pingmode != PING_FLOOD)
		(void) ping_alarm(0);

	for (;;)
	{
		/* send flood until reply comes in */
		if (pingmode == PING_FLOOD)
			(void) ping_alarm(0);

		/* wait for reply packet, or alarm */
		cc = recv_ping();

		/* process input packet */
		if (check_ping(ipacket, cc))
			gotone = TRUE;

		/* terminate when alive in quick mode */
		if (gotone && quick)
			npackets = ntransmitted;

		/* flush input queue in broadcast mode */
		if (broadcast)
			continue;

		/* terminate if we got enough */
		if (npackets > 0 && pingstats.rcvd >= npackets)
		{
			(void) finish(0);
			/*NOTREACHED*/
		}

		/* schedule immediate next event if necessary */
		if (gotone && fastping)
			(void) ping_alarm(0);
	}
	/*NOTREACHED*/
}

/*
** PING_ALARM -- Handle an alarm
** -----------------------------
**
**	This routine causes another PING to be transmitted, and
**	schedules another SIGALRM at the proper period from now.
**	In flood mode, a flood of pings is transmitted at short
**	intervals. It returns as soon as a reponse is available.
**	Shutdown is initiated in all modes if appropriate.
** 
** 	Our sense of time will slowly skew (ie, packets will not
**	be launched exactly at 1-second intervals). This does not
**	affect the quality of the delay and loss statistics.
*/

sigtype_t
ping_alarm(sig)
int sig;				/* nonzero if this was an alarm */
{
	int waittime;			/* timeout until shutdown */
	int ready;

/*
 * Issue warning if nothing was received within the timeout period.
 */
	if (sig == SIGALRM && !gotone)
	{
		if (!quiet)
		{
			if (pingmode == PING_CISCO)
				c_put(".");
			else if (pingmode == PING_NORMAL)
				printf("no reply from %s\n", hostname);
		}
	}

	/* reset for next try */
	gotone = FALSE;

/*
 * Retry until the specified packet limit is reached.
 */
restart:
	if (npackets == 0 || ntransmitted < npackets)
	{
		/* must send another packet */
		send_ping();

		if (pingmode == PING_FLOOD)
		{
			/* wait only a very short time */
			ready = wait_ping(floodtime);
			if (ready == 0)
				goto restart;
		}
		else
		{
			/* schedule next alarm */
			(void) signal(SIGALRM, ping_alarm);
			(void) alarm((unsigned)timeout);
		}
	}
	else if (broadcast || (pingstats.rcvd < ntransmitted))
	{
		/* must pickup outstanding packets */
		flushing = TRUE;

		if (pingmode == PING_CISCO)
		{
			/* standard timeout */
			waittime = timeout;
		}
		else if (pingstats.rcvd > 0)
		{
			/* make educated guess */
			waittime = timeout + (pingstats.rttmax / 1000);
		}
		else
		{
			/* maximum timeout */
			waittime = MAXWAIT;
		}

		/* schedule shutdown */
		(void) signal(SIGALRM, finish);
		(void) alarm((unsigned)waittime);
	}
	else
	{
		(void) finish(sig);
		/*NOTREACHED*/
	}
}

/*
** SEND_PING -- Send a ping packet
** -------------------------------
**
**	Compose and transmit an ICMP ECHO request packet. The IP packet
**	will be added on by the kernel. The ID field is our UNIX process ID,
**	and the sequence number is an ascending integer. The first 8 bytes
**	of the data portion are used to hold a UNIX timeval struct in host
**	byte-order, to compute the round-trip time.
*/

void
send_ping()
{
	struct icmp *icp;		/* icmp packet */
	struct timeval *sendtime;	/* time when transmitted */
	int len;			/* size of icmp output packet */
	int cc;				/* size actually transmitted */

	/* store current time in output packet */
	if (timing)
	{
		sendtime = (struct timeval *)&opacket[HDRLEN];
		(void) gettimeofday(sendtime, (struct timezone *)NULL);
	}

	/* total size of output packet */
	len = HDRLEN + datalen;

	/* construct icmp header */
	icp = (struct icmp *)opacket;

	icp->icmp_type  = request;
	icp->icmp_code  = 0;
	icp->icmp_cksum = 0;
	icp->icmp_id    = ident;
	icp->icmp_seq   = ntransmitted++;
	icp->icmp_cksum = in_cksum((u_short *)icp, len);

	/* clear this packet in stats table */
	CLRBIT(icp->icmp_seq);

	/* transmit the output packet */
	cc = sendto(sock, (char *)opacket, len, 0, &toaddr, sizeof(toaddr));
	if (cc < 0 || cc != len)
	{
		if (!quiet && pingmode != PING_NORMAL)
			(void) fprintf(stderr, "\n");
		if (cc < 0)
			perror("sendto");
		else
			error("sendto: truncated packet to %s", hostname);

#ifdef EMSGSIZE
		/* message too long */
		if (errno == EMSGSIZE)
			exit(EX_DATAERR);
#endif /*EMSGSIZE*/

		/* don't retry in quick mode */
		if (quick)
			exit(EX_UNAVAILABLE);
	}

	/* display marker if appropriate */
	if (!quiet)
	{
		if (pingmode == PING_FLOOD)
			(void) write(STDOUT, ".", 1);
	}
}

/*
** RECV_PING -- Input a ping packet
** --------------------------------
**
**	Read a new input packet. Normally we wait here for the arrival
**	of a packet. We may be interrupted by an alarm if no packet
**	came in during the current timeout period. In flood ping mode
**	there are no alarms, and the arrival is indicated via select().
**
**	Returns:
**		The number of bytes in the input packet.
*/

int
recv_ping()
{
	int fromlen;			/* size of address buffer */
	int len;			/* size of input packet buffer */
	int cc;				/* size actually read */

restart:
	len = sizeof(ipacket);
	fromlen = sizeof(fromaddr);
	cc = recvfrom(sock, (char *)ipacket, len, 0, &fromaddr, &fromlen);
	if (cc <= 0)
	{
		/* shouldn't happen */
		if (cc == 0)
			errno = ECONNRESET;

		/* interrupt -- restart */
		if (errno == EINTR)
			goto restart;

		if (!quiet && pingmode != PING_NORMAL)
			(void) fprintf(stderr, "\n");
		perror("recvfrom");

		(void) close(sock);
		exit(EX_OSERR);
	}

	return(cc);
}

/*
** WAIT_PING -- Wait for a ping packet
** -----------------------------------
**
**	Wait for the arrival of a new input packet via select().
**	This is used only in flood ping mode. A very short timeout
**	is used, so that a new ping can be sent quickly.
**
**	Returns:
**		Nonzero if input is available.
**		Zero in case of timeout.
*/

int
wait_ping(millisecs)
int millisecs;				/* timeout value in millisecs */
{
	struct timeval timer;
	fd_set fds;
	int ready;

	timer.tv_sec = 0;
	timer.tv_usec = millisecs * 1000;

restart:
	/* FD_ZERO(&fds); */
	bzero((char *)&fds, sizeof(fds));
	FD_SET(sock, &fds);

	/* wait only a very short time */
	ready = select(FD_SETSIZE, &fds, (fd_set *)0, (fd_set *)0, &timer);
	if (ready < 0)
	{
		/* interrupt -- restart */
		if (errno == EINTR)
			goto restart;

		if (!quiet && pingmode != PING_NORMAL)
			(void) fprintf(stderr, "\n");
		perror("select");

		(void) close(sock);
		exit(EX_OSERR);
	}

	/* indicate availability of packet */
	return(ready);
}

/*
** CHECK_PING -- Process incoming icmp packet
** ------------------------------------------
**
**	Print out the packet, if it came from us. This logic is necessary
**	because ALL readers of the ICMP socket get a copy of ALL ICMP packets
**	which arrive ('tis only fair). This permits multiple copies of this
**	program to be run without having intermingled output (or statistics!).
**
**	Returns:
**		TRUE if this is a valid response packet.
**		FALSE otherwise.
*/

bool
check_ping(buf, cc)
u_char *buf;				/* input packet */
int cc;					/* size of input packet */
{
	struct ip *ip;
	struct icmp *icp;
	int iphdrlen;			/* ip header length of input packet */
	struct timeval tv;		/* elapsed time */
	struct timeval *recvtime;	/* time input packet is received */
	struct timeval *sendtime;	/* time as stored in icmp packet */
	time_t rtt;			/* round-trip time in millisecs */
	bool duplicate;			/* set if this is a duplicate reply */

	/* pickup current time */
	if (timing)
	{
		recvtime = &tv;
		(void) gettimeofday(recvtime, (struct timezone *)NULL);
	}

/*
 * The input packet contains the ip header in front of the icmp packet.
 * Make sure the packet contains the icmp header after the ip header.
 */
	ip = (struct ip *)buf;
	iphdrlen = ip->ip_hl << 2;

	if (ip->ip_p != IPPROTO_ICMP)
	{
		if (verbose)
		{
			printf("\n%d bytes from %s: (not icmp)\n",
				cc, pr_addr(from->sin_addr));

			/* dump the ip packet */
			pr_ippkt(ip);
		}
		return(FALSE);
	}

	if (cc < iphdrlen + ICMP_MINLEN)
	{
		if (verbose)
		{
			printf("\n%d bytes from %s: (too short)\n",
				cc, pr_addr(from->sin_addr));

			/* dump the ip packet */
			pr_ippkt(ip);
		}
		return(FALSE);
	}

	/* move to the icmp packet */
	icp = (struct icmp *)(buf + iphdrlen);
	cc -= iphdrlen;

/*
 * Check whether we got the expected reply type.
 */
	if (icp->icmp_type != reply)
	{
		if (verbose)
		{
			printf("\n%d bytes from %s: (wrong type) ",
				cc, pr_addr(from->sin_addr));

			/* dump the icmp packet */
			pr_icmph(icp);
		}
		return(FALSE);
	}

/*
 * Check whether it belongs to us.
 */
	if (icp->icmp_id != ident)
	{
		if (verbose)
		{
			printf("\n%d bytes from %s: (wrong ident) ",
				cc, pr_addr(from->sin_addr));

			/* dump the icmp packet */
			pr_icmph(icp);
		}
		return(FALSE);
	}

	/* XXX should compare the patterns to make sure the packets match */
	if (cc != HDRLEN + datalen)
	{
		if (verbose)
		{
			printf("\n%d bytes from %s: (wrong size) ",
				cc, pr_addr(from->sin_addr));

			/* dump the icmp packet */
			pr_icmph(icp);
		}
		return(FALSE);
	}

/*
 * Check for duplicate reply.
 */
	if (broadcast)
		duplicate = sameaddr(from, to);
	else
		duplicate = TSTBIT(icp->icmp_seq);

	if (duplicate)
		pingstats.dupl++;
	else
		pingstats.rcvd++;

	if (!duplicate)
		SETBIT(icp->icmp_seq);

	if (broadcast)
		update_hosts(from->sin_addr);

/*
 * Compute round-trip time.
 */
	if (timing)
	{
		sendtime = (struct timeval *)icp->icmp_data;
		tvsub(recvtime, sendtime);
		rtt = (tv.tv_sec*1000) + ((tv.tv_usec+500)/1000);
	}

	if (timing && !duplicate)
		record_stats(&pingstats, rtt);

/*
 * The options buffer may be present at the end of the ip header.
 */
#ifdef IP_OPTIONS
	if ((traceroute || looseroute) && (iphdrlen > IPHDRSZ))
	{
		u_char *ipopt;		/* start of options buffer */
		int ipoptlen;		/* total size of options buffer */

		ipopt = (u_char *)ip + IPHDRSZ;
		ipoptlen = iphdrlen - IPHDRSZ;

		if (ipoptlen > 0 && ipoptlen <= MAX_IPOPTLEN)
		{
			register optstr *opt;

			/* locate identical data in current chain */
			for (opt = optchain; opt != NULL; opt = opt->next)
			{
				if (ipoptlen != opt->ipoptlen)
					continue;

				if (bcmp((char *)ipopt, (char *)opt->ipopt, ipoptlen) == 0)
					break;
			}

			/* allocate new entry if this is new data */
			if (opt == NULL)
			{
				opt = (optstr *)malloc(sizeof(optstr));
				if (opt != NULL)
				{
					zero_stats(&opt->stats);
					bcopy((char *)ipopt, (char *)opt->ipopt, ipoptlen);
					opt->ipoptlen = ipoptlen;
					opt->next = optchain;
					optchain = opt;
				}
			}

			/* update statistics for this entry */
			if (opt != NULL)
			{
				if (duplicate)
					opt->stats.dupl++;
				else
					opt->stats.rcvd++;

				if (timing && !duplicate)
					record_stats(&opt->stats, rtt);
			}
		}
	}
#endif /*IP_OPTIONS*/

/*
 * Print appropriate reponse.
 */
	if (!quiet)
	{
		if (pingmode == PING_FLOOD)
		{
			if (duplicate)
				(void) write(STDOUT, "*", 1);
			else if (flushing)
				(void) write(STDOUT, "!", 1);
			else if (broadcast)
				(void) write(STDOUT, "!", 1);
			else
				(void) write(STDOUT, "\b!\b", 3);
		}
		else if (pingmode == PING_CISCO)
		{
			if (duplicate)
				c_put("*");
			else
				c_put("!");
		}
		else
		{
			printf("%d bytes from %s: seq=%d ttl=%d",
				cc, inet_ntoa(from->sin_addr),
				(int)icp->icmp_seq, (int)ip->ip_ttl);
			if (timing)
				printf(" time=%ld ms", rtt);
			if (duplicate)
				printf(", duplicate");
			printf(".\n");
		}
	}

/*
 * Print options buffer, if appropriate.
 */
#ifdef IP_OPTIONS
	if (verbose > 1 && iphdrlen > IPHDRSZ)
	{
		u_char *ipopt;		/* start of options buffer */
		int ipoptlen;		/* total size of options buffer */

		ipopt = (u_char *)ip + IPHDRSZ;
		ipoptlen = iphdrlen - IPHDRSZ;

		pr_options(ipopt, ipoptlen);
	}
#endif /*IP_OPTIONS*/

	/* return valid packet indication */
	return(!duplicate);
}

/*
** PREFINISH -- Prepare for shutdown
** ---------------------------------
**
**	On the first SIGINT, allow any outstanding packets to dribble in.
**	In quick mode, or if nothing has been sent at all, we can finish.
**	If nothing has been received yet, we assume that we are dealing
**	with a dead target and we quit as well. In broadcast mode we are
**	always waiting for more to come.
*/

sigtype_t
prefinish(sig)
int sig;				/* nonzero on interrupt */
{
	/* quit now if nothing transmitted or none outstanding */
	if (quick || ntransmitted == 0 || pingstats.rcvd == 0 ||
	    (!broadcast && (pingstats.rcvd >= ntransmitted)))
	{
		(void) finish(sig);
		/*NOTREACHED*/
	}

	/* final shutdown on next interrupt */
	(void) signal(SIGINT, finish);

	/* let the normal limit work */
	npackets = ntransmitted;

	printf("\n---- Waiting for outstanding packets ----\n");
	column = 0;
}

/*
** FINISH -- Clean up and terminate
** --------------------------------
**
**	Print out statistics summary, and terminate.
**
**	Exit status:
**		EX_OK if at least one valid response was received.
**		EX_UNAVAILABLE if there were none.
*/

sigtype_t
finish(sig)
int sig;				/* nonzero on interrupt */
{
	int exitstat;			/* ping exit status code */

/*
 * Reset state.
 */
	/* no more background action */
	(void) signal(SIGALRM, SIG_IGN);
	(void) alarm((unsigned)0);

	/* and no more special interrupt handling */
	(void) signal(SIGINT, SIG_DFL);

	/* indicate whether we got any valid response */
	exitstat = (pingstats.rcvd > 0) ? EX_OK : EX_UNAVAILABLE;

/*
 * Show status if no statistics required.
 */
	if (quick)
	{
		if (sig == SIGINT)
			printf("\n");
		if (pingstats.rcvd > 0)
			printf("%s is alive\n", hostname);
		else if (ntransmitted > 0)
			printf("no reply from %s\n", hostname);

		exit(exitstat);
		/*NOTREACHED*/
	}

/*
 * Print general statistics.
 */
	printf("\n---- %s PING Statistics ----\n",
		pr_addr(to->sin_addr));
	printf("%d packet%s transmitted",
		ntransmitted, plural(ntransmitted));
	printf(", %d packet%s received",
		pingstats.rcvd, plural(pingstats.rcvd));
	if (pingstats.dupl > 0)
		printf(", %d duplicate%s",
		pingstats.dupl, plural(pingstats.dupl));
	if (broadcast)
		printf(", broadcast");
	else if (pingstats.rcvd > ntransmitted)
		printf(" -- somebody's printing up packets!");
	else if (ntransmitted > 0)
	{
		int missed = ntransmitted - pingstats.rcvd;
		double loss = 100 * (double)missed / (double)ntransmitted;

		if (pingstats.rcvd == 0)
			printf(", %d%% packet loss", 100);
		else
			printf(", %.2g%% packet loss", loss);
	}
	printf("\n");

/*
 * Print timing statistics, if appropriate.
 */
	if (timing)
		print_stats(&pingstats);

/*
 * Print responding hosts in broadcast mode.
 */
	if (broadcast)
	{
		register hostinfo *h;
		int hostcount;

		hostcount = 0;
		for (h = hostchain; h != NULL; h = h->next)
			hostcount++;

		if (hostcount == 0)
			printf("No hosts responded\n");
		else
			printf("%d different host%s responded\n",
				hostcount, plural(hostcount));

		for (h = hostchain; h != NULL; h = h->next)
		{
			printf("\t%d packet%s", h->rcvd, plural(h->rcvd));
			if (ntransmitted > 0)
			{
				double frac = 100 * (double)h->rcvd /
						    (double)ntransmitted;

				if (h->rcvd == ntransmitted)
					printf(" (%d%%)", 100);
				else if (h->rcvd < ntransmitted)
					printf(" (%.2g%%)", frac);
			}
			printf(" from %s\n", pr_addr(h->inaddr));
		}
	}

/*
 * Print recorded routes, if available.
 */
#ifdef IP_OPTIONS
	if (traceroute || looseroute)
	{
		register optstr *opt;
		int optcount;

		optcount = 0;
		for (opt = optchain; opt != NULL; opt = opt->next)
			optcount++;

		if (optcount == 0)
			printf("No packets with routing info received\n");
		else if (optcount > 1)
			printf("%d different route%s received\n",
				optcount, plural(optcount));

		for (opt = optchain; opt != NULL; opt = opt->next)
		{
			printf("\n%d packet%s",
				opt->stats.rcvd, plural(opt->stats.rcvd));
			if (pingstats.rcvd > 0)
			{
				double frac = 100 * (double)opt->stats.rcvd /
						    (double)pingstats.rcvd;

				if (opt->stats.rcvd == pingstats.rcvd)
					printf(" (%d%%)", 100);
				else
					printf(" (%.2g%%)", frac);
			}
			if (opt->stats.dupl > 0)
				printf(" %d duplicate%s",
				opt->stats.dupl, plural(opt->stats.dupl));
			printf(" via:\n");
			pr_options(opt->ipopt, opt->ipoptlen);
			if (timing)
				print_stats(&opt->stats);
		}
	}
#endif /*IP_OPTIONS*/

/*
 * Print missing responses (accurate only if we didn't wrap).
 */
	if (printmiss)
	{
		register int response;
		int missing, maxmissed;

		maxmissed = ntransmitted;
		if (maxmissed > BITMAPSIZE)
			maxmissed = BITMAPSIZE;

		missing = 0;
		for (response = 0; response < maxmissed; response++)
			if (!TSTBIT(response))
				missing++;

		printf("%d missed response%s\n", missing, plural(missing));
		if (missing > 0 && missing < maxmissed)
		{
			missing = 0;
			for (response = 0; response < maxmissed; response++)
			{
				if (!TSTBIT(response))
				{
					missing++;
					printf("%3d ", response);
					if ((missing % 15) == 0)
						printf("\n");
				}
			}

			if ((missing % 15) != 0)
				printf("\n");
		}
	}

/*
 * Final termination.
 */
	exit(exitstat);
	/*NOTREACHED*/
}

/*
** PR_ICMPH -- Print a descriptive string about an ICMP header
** -----------------------------------------------------------
*/

void
pr_icmph(icp)
struct icmp *icp;			/* icmp buffer */
{
	switch (icp->icmp_type)
	{
	    case ICMP_ECHOREPLY:
		printf("Echo Reply\n");
		/* XXX ID + Seq + Data */
		break;

	    case ICMP_UNREACH:
		switch (icp->icmp_code)
		{
		    case ICMP_UNREACH_NET:
			printf("Dest Net Unreachable\n");
			break;
		    case ICMP_UNREACH_HOST:
			printf("Dest Host Unreachable\n");
			break;
		    case ICMP_UNREACH_PROTOCOL:
			printf("Dest Protocol Unreachable\n");
			break;
		    case ICMP_UNREACH_PORT:
			printf("Dest Port Unreachable\n");
			break;
		    case ICMP_UNREACH_NEEDFRAG:
			printf("Frag needed and DF set\n");
			break;
		    case ICMP_UNREACH_SRCFAIL:
			printf("Source Route Failed\n");
			break;
		    default:
			printf("Dest Unreachable, Bad Code: %d\n", icp->icmp_code);
			break;
		}
		pr_ippkt((struct ip *)icp->icmp_data);
		break;

	    case ICMP_SOURCEQUENCH:
		printf("Source Quench\n");
		pr_ippkt((struct ip *)icp->icmp_data);
		break;

	    case ICMP_REDIRECT:
		switch (icp->icmp_code)
		{
		    case ICMP_REDIRECT_NET:
			printf("Redirect Net");
			break;
		    case ICMP_REDIRECT_HOST:
			printf("Redirect Host");
			break;
		    case ICMP_REDIRECT_TOSNET:
			printf("Redirect TOS and Net");
			break;
		    case ICMP_REDIRECT_TOSHOST:
			printf("Redirect TOS and Host");
			break;
		    default:
			printf("Redirect, Bad Code: %d", icp->icmp_code);
			break;
		}
		printf(" (New: %s)\n", inet_ntoa(icp->icmp_gwaddr));
		pr_ippkt((struct ip *)icp->icmp_data);
		break;

	    case ICMP_ECHO:
		printf("Echo Request\n");
		/* XXX ID + Seq + Data */
		break;

	    case ICMP_TIMXCEED:
		switch (icp->icmp_code)
		{
		    case ICMP_TIMXCEED_INTRANS:
			printf("Time to live exceeded\n");
			break;
		    case ICMP_TIMXCEED_REASS:
			printf("Frag reassembly time exceeded\n");
			break;
		    default:
			printf("Time exceeded, Bad Code: %d\n", icp->icmp_code);
			break;
		}
		pr_ippkt((struct ip *)icp->icmp_data);
		break;

	    case ICMP_PARAMPROB:
		printf("Parameter Problem");
		printf(" (Pointer: 0x%02x)\n", icp->icmp_pptr);
		pr_ippkt((struct ip *)icp->icmp_data);
		break;

	    case ICMP_TSTAMP:
		printf("Timestamp\n");
		/* XXX ID + Seq + 3 timestamps */
		break;

	    case ICMP_TSTAMPREPLY:
		printf("Timestamp Reply\n");
		/* XXX ID + Seq + 3 timestamps */
		break;

	    case ICMP_IREQ:
		printf("Information Request\n");
		/* XXX ID + Seq */
		break;

	    case ICMP_IREQREPLY:
		printf("Information Reply\n");
		/* XXX ID + Seq */
		break;

	    case ICMP_MASKREQ:
		printf("Address Mask Request\n");
		break;

	    case ICMP_MASKREPLY:
		printf("Address Mask Reply\n");
		break;

	    default:
		printf("Bad ICMP type: %d\n", icp->icmp_type);
		break;
	}
}

/*
** PR_IPPKT -- Dump some info on a returned (via ICMP) IP packet
** -------------------------------------------------------------
*/

void
pr_ippkt(ip)
struct ip *ip;				/* returned ip packet buffer */
{
	int iphdrlen;			/* total size of ip packet */
	struct tcphdr *tcp;		/* start of tcp packet */
	struct udphdr *udp;		/* start of udp packet */

/*
 * Print ip header itself.
 */
	pr_iph(ip);

/*
 * Plus extra info for certain protocols.
 */
	if (ip->ip_p == IPPROTO_TCP)
	{
		iphdrlen = ip->ip_hl << 2;
		tcp = (struct tcphdr *)((u_char *)ip + iphdrlen);

		printf("TCP: ");
		printf("from port %s", pr_port("tcp", tcp->th_sport));
		printf(", to port %s", pr_port("tcp", tcp->th_dport));
		printf("\n");
	}
	else if (ip->ip_p == IPPROTO_UDP)
	{
		iphdrlen = ip->ip_hl << 2;
		udp = (struct udphdr *)((u_char *)ip + iphdrlen);

		printf("UDP: ");
		printf("from port %s", pr_port("udp", udp->uh_sport));
		printf(", to port %s", pr_port("udp", udp->uh_dport));
		printf("\n");
	}
}

/*
** PR_IPH -- Print an IP header with options
** -----------------------------------------
*/

void
pr_iph(ip)
struct ip *ip;				/* returned ip packet buffer */
{
	int iphdrlen;			/* total size of ip packet */

/*
 * Dump the ip header.
 */
	printf("VR HL TOS  LEN   ID FLG  OFF TTL PRO  CKS SRC             DST\n");
	printf("%2d %2d", ip->ip_v, ip->ip_hl);
	printf("  %02x",  ip->ip_tos);
	printf(" %4d",    (int)ip->ip_len);
	printf(" %04x",   ip->ip_id);
	printf("   %01x", ((ip->ip_off) & 0xE000) >> 13);
	printf(" %04x",   ((ip->ip_off) & 0x1FFF));
	printf(" %3d",    (int)ip->ip_ttl);
	printf(" %3d",    (int)ip->ip_p);
	printf(" %04x",   ip->ip_sum);
	printf(" %-15s",  inet_ntoa(ip->ip_src));
	printf(" %-15s",  inet_ntoa(ip->ip_dst));
	printf("\n");

/*
 * Dump option bytes.
 */
	iphdrlen = ip->ip_hl << 2;
	if (iphdrlen > IPHDRSZ)
	{
		register int i;
		u_char *ipopt;		/* address of options buffer */
		int ipoptlen;		/* total size of options buffer */

		ipopt = (u_char *)ip + IPHDRSZ;
		ipoptlen = iphdrlen - IPHDRSZ;

		printf("IPOPT:");
		for (i = 0; i < ipoptlen; i++)
			printf(" %02x", ipopt[i]);
		printf("\n");
#ifdef IP_OPTIONS
		pr_options(ipopt, ipoptlen);
#endif /*IP_OPTIONS*/
	}
}

/*
** PR_OPTIONS -- Print ip options data
** -----------------------------------
*/

#ifdef IP_OPTIONS

void
pr_options(ipopt, ipoptlen)
u_char *ipopt;				/* address of options buffer */
int ipoptlen;				/* total size of options buffer */
{
	int optval;			/* option value */
	int optlen;			/* size of this option */

	while (ipoptlen > 0)
	{
		optval = ipopt[IPOPT_OPTVAL];
		optlen = ipopt[IPOPT_OLEN];

		switch (optval)
		{
		    case IPOPT_EOL :
			optlen = 0;
			break;

		    case IPOPT_NOP :
			optlen = 1;
			break;

		    case IPOPT_RR :
			printf("<%s>\n", "recorded route");
			pr_route(ipopt);
			break;

		    case IPOPT_LSRR :
			printf("<%s>\n", "loose source route");
			pr_route(ipopt);
			break;

		    case IPOPT_SSRR :
			printf("<%s>\n", "strict source route");
			pr_route(ipopt);
			break;

		    case IPOPT_TS :
			printf("<%s>\n", "time stamp");
			break;

		    case IPOPT_SECURITY :
			printf("<%s>\n", "security");
			break;

		    case IPOPT_SATID :
			printf("<%s>\n", "stream id");
			break;

		    default:
			printf("<option %d, length %d>\n", optval, optlen);
			break;
		}

		if (optlen == 0)
			optlen = ipoptlen;
		ipopt += optlen;
		ipoptlen -= optlen;
	}
}

#endif /*IP_OPTIONS*/

/*
** PR_ROUTE -- Print ip route data
** -------------------------------
*/

#ifdef IP_OPTIONS

void
pr_route(ipopt)
u_char *ipopt;				/* start of current option */
{
	int optval;			/* option value */
	int optlen;			/* size of this option */
	u_char *optptr;			/* pointer to option data */
	int offset;			/* length of option data */

	optval = ipopt[IPOPT_OPTVAL];
	optlen = ipopt[IPOPT_OLEN];
	offset = ipopt[IPOPT_OFFSET];

	optptr = &ipopt[IPOPT_HDRLEN];
	optlen -= IPOPT_HDRLEN;

	while (offset > IPOPT_MINOFF)
	{
		struct in_addr inaddr;

		bcopy((char *)optptr, (char *)&inaddr, INADDRSZ);
		printf("\t%s\n", pr_addr(inaddr));

		optptr += INADDRSZ;
		offset -= INADDRSZ;
		optlen -= INADDRSZ;
	}

	/* buffer full -- perhaps route not complete */
	if (optval == IPOPT_RR && optlen == 0)
		printf("\t%s\n", "...");
}

#endif /*IP_OPTIONS*/

/*
** PR_ADDR -- Return a host name and/or dotted quad
** ------------------------------------------------
*/

char *
pr_addr(inaddr)
struct in_addr inaddr;			/* IP address */
{
	static char buf[BUFSIZ];
	struct hostent *hp;

	if (numeric)
		hp = NULL;
	else
		hp = gethostbyaddr((char *)&inaddr, INADDRSZ, AF_INET);

	if (hp == NULL)
		(void) sprintf(buf, "%s", inet_ntoa(inaddr));
	else
		(void) sprintf(buf, "%s (%s)", hp->h_name, inet_ntoa(inaddr));
	return(buf);
}

/*
** PR_PORT -- Return a service port name
** -------------------------------------
*/

char *
pr_port(protocol, port)
char *protocol;				/* the protocol used */
u_short port;				/* port number in network order */
{
	struct servent *service;
	static char buf[BUFSIZ];

	service = getservbyport((int)port, protocol);
	if (service != NULL)
		(void) sprintf(buf, "%s", service->s_name);
	else
		(void) sprintf(buf, "%d", (int)ntohs(port));
	return(buf);
}

/*
** PRINT_STATS -- Print round_trip statistics
** ------------------------------------------
*/

void
print_stats(sp)
stats *sp;				/* statistics buffer */
{
	double rttavg;			/* average round-trip time */

	if (sp->rcvd > 0)
	{
		rttavg = sp->rttsum / sp->rcvd;
		printf("round-trip (ms)\tmin/avg/max = %ld/%.0f/%ld\n",
			sp->rttmin, rttavg, sp->rttmax);
	}
}

/*
** RECORD_STATS -- Update round-trip statistics
** --------------------------------------------
*/

void
record_stats(sp, rtt)
stats *sp;				/* statistics buffer */
time_t rtt;				/* round-trip time */
{
	if (rtt < sp->rttmin)
		sp->rttmin = rtt;

	if (rtt > sp->rttmax)
		sp->rttmax = rtt;

	sp->rttsum += (double)rtt;
}

/*
** ZERO_STATS -- Clear out round-trip statistics
** ---------------------------------------------
*/

void
zero_stats(sp)
stats *sp;				/* statistics buffer */
{
	sp->rcvd = 0;
	sp->dupl = 0;
	sp->rttmin = VERY_LONG;
	sp->rttmax = 0;
	sp->rttsum = 0.0;
}

/*
** UPDATE_HOSTS -- Register host during broadcast
** ----------------------------------------------
*/

void
update_hosts(inaddr)
struct in_addr inaddr;			/* IP address */
{
	register hostinfo *h;
	register hostinfo **hp;

	/* check if this address is known */
	for (h = hostchain; h != NULL; h = h->next)
		if (h->inaddr.s_addr == inaddr.s_addr)
			break;

	/* allocate new entry if unknown */
	if (h == NULL)
	{
		/* sort according to ascending IP address */
		for (hp = &hostchain; (h = *hp) != NULL; hp = &h->next)
			if (ntohl(inaddr.s_addr) < ntohl(h->inaddr.s_addr))
				break;

		h = (hostinfo *)malloc(sizeof(hostinfo));
		if (h != NULL)
		{
			h->inaddr = inaddr;
			h->rcvd = 0;
			h->next = *hp;
			*hp = h;
		}
	}

	/* update packet count */
	if (h != NULL)
		h->rcvd++;
}

/*
** C_PUT -- Print markers in Cisco style
** -------------------------------------
*/

void
c_put(str)
char *str;				/* marker to print (first char only) */
{
	if (column >= 79)
	{
		(void) write(STDOUT, "\n", 1);
		column = 0;
	}
	(void) write(STDOUT, str, 1);
	column++;
}

/*
** TVSUB -- Subtract two timeval structs to compute delay
** ------------------------------------------------------
**
**	The difference of the two values is stored back into the first.
*/

void
tvsub(t2, t1)
struct timeval *t2;			/* ending time */
struct timeval *t1;			/* starting time */
{
	t2->tv_usec -= t1->tv_usec;
	if (t2->tv_usec < 0)
	{
		t2->tv_sec--;
		t2->tv_usec += 1000000;
	}

	if (t2->tv_sec < t1->tv_sec)
	{
		t2->tv_sec = 0;
		t2->tv_usec = 0;
	}
	else
		t2->tv_sec -= t1->tv_sec;
}

/*
** IN_CKSUM -- Compute checksum for IP packets
** -------------------------------------------
**
**	The complete packet must have been constructed.
**	The checksum field to be computed must be zero.
**
**	Returns:
**		Computed checksum.
*/

#ifdef obsolete

u_short
in_cksum(buf, len)
u_short *buf;				/* start of packet */
int len;				/* length of packet in bytes */
{
	register u_short *w = buf;	/* address of next 16-bit word */
	register int nleft = len;	/* remaining 16-bit words */
	register int sum = 0;		/* 32-bit accumulator */
	u_short answer = 0;

/*
 * Our algorithm is simple, using a 32 bit accumulator (sum),
 * we add sequential 16 bit words to it, and at the end, fold back
 * all the carry bits from the top 16 bits into the lower 16 bits.
 */
	while (nleft > 1)
	{
		sum += *w++;
		nleft -= 2;
	}

	if (nleft == 1)
	{
		*(u_char *)(&answer) = *(u_char *)w;
		sum += answer;
	}

/*
 * Add back carry outs from top 16 bits to low 16 bits.
 */
	sum = (sum >> 16) + (sum & 0xFFFF);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

#endif /*obsolete*/

/*
** IN_CKSUM -- Compute checksum for IP packets
** -------------------------------------------
**
**	The complete packet must have been constructed.
**	The checksum field to be computed must be zero.
**
**	Returns:
**		Computed checksum.
*/

u_short
in_cksum(buf, len)
u_short *buf;				/* start of packet */
int len;				/* length of packet in bytes */
{
	register u_char *b = (u_char *)buf;
	register int n = len;		/* remaining 16-bit words */
	register int sum = 0;		/* 32-bit accumulator */
	u_short answer;

/*
 * Our algorithm is simple, using a 32 bit accumulator (sum),
 * we add sequential 16 bit words to it, and at the end, fold back
 * all the carry bits from the top 16 bits into the lower 16 bits.
 */
	while (n > 1)
	{
		answer = (b[0] << 8) | b[1];
		sum += answer;
		b += 2; n -= 2;
	}

	if (n == 1)
	{
		answer = (b[0] << 8);
		sum += answer;
	}

/*
 * Add back carry outs from top 16 bits to low 16 bits.
 */
	sum = (sum >> 16) + (sum & 0xFFFF);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum & 0xFFFF;			/* truncate to 16 bits */
	answer = htons(answer);			/* correct order */
	return(answer);
}

/*
** BCAST_ADDR -- Check if this represents a broadcast address
** ----------------------------------------------------------
**
**	Returns:
**		TRUE if this is a broadcast address.
**		FALSE otherwise.
*/

#define	CLASSA(a)	(((a) & (ipaddr_t)0x80000000) == (ipaddr_t)0x00000000)
#define	CLASSB(a)	(((a) & (ipaddr_t)0xC0000000) == (ipaddr_t)0x80000000)
#define	CLASSC(a)	(((a) & (ipaddr_t)0xE0000000) == (ipaddr_t)0xC0000000)

#define	CLASSD(a)	(((a) & (ipaddr_t)0xF0000000) == (ipaddr_t)0xE0000000)
#define	CLASSE(a)	(((a) & (ipaddr_t)0xF0000000) == (ipaddr_t)0xF0000000)
#define	CLASSL(a)	(((a) & (ipaddr_t)0xFF000000) == (ipaddr_t)0x7F000000)

#define	CLASSA_MASK	(ipaddr_t)0x00FFFFFF
#define	CLASSB_MASK	(ipaddr_t)0x0000FFFF
#define	CLASSC_MASK	(ipaddr_t)0x000000FF

bool
bcast_addr(inaddr)
struct in_addr inaddr;			/* IP address */
{
	register ipaddr_t address = ntohl(inaddr.s_addr);
	register ipaddr_t hostmask;

	if (CLASSL(address))
		return(FALSE);		/* loopback */
	else if (CLASSA(address))
		hostmask = CLASSA_MASK;
	else if (CLASSB(address))
		hostmask = CLASSB_MASK;
	else if (CLASSC(address))
		hostmask = CLASSC_MASK;
	else if (CLASSD(address))
		return(TRUE);		/* multicast */
	else
		return(FALSE);		/* reserved */

	/* mask host part */
	address &= hostmask;

	/* must be all zeroes or all ones */
	if (address == 0 || address == hostmask)
		return(TRUE);

	/* XXX should check for subnets here */
	return(FALSE);
}

/*
** ITOA -- Convert integer value to ascii string
** ---------------------------------------------
**
**	Returns:
**		Pointer to string.
*/

char *
itoa(n)
int n;					/* value to convert */
{
	static char buf[30];

	(void) sprintf(buf, "%d", n);
	return(buf);
}
